"""
PynamoDB attributes
"""
import base64
import calendar
import collections.abc
import json
import time
import warnings
from base64 import b64encode, b64decode
from copy import deepcopy
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from inspect import getfullargspec
from inspect import getmembers
from typing import Any, Callable, Dict, Generic, List, Mapping, Optional, TypeVar, Type, Union, Set, overload, Iterable
from typing import TYPE_CHECKING

from pynamodb._util import attr_value_to_simple_dict
from pynamodb._util import bin_decode_attr
from pynamodb._util import bin_encode_attr
from pynamodb._util import simple_dict_to_attr_value
from pynamodb.constants import BINARY
from pynamodb.constants import BINARY_SET
from pynamodb.constants import BOOLEAN
from pynamodb.constants import DATETIME_FORMAT
from pynamodb.constants import LIST
from pynamodb.constants import MAP
from pynamodb.constants import NULL
from pynamodb.constants import NUMBER
from pynamodb.constants import NUMBER_SET
from pynamodb.constants import STRING
from pynamodb.constants import STRING_SET
from pynamodb.exceptions import AttributeDeserializationError
from pynamodb.exceptions import AttributeNullError
from pynamodb.expressions.operand import Path


if TYPE_CHECKING:
    from pynamodb.expressions.condition import (
        BeginsWith, Between, Comparison, Contains, NotExists, Exists, In
    )
    from pynamodb.expressions.operand import (
        _Decrement, _IfNotExists, _Increment, _ListAppend
    )
    from pynamodb.expressions.update import (
        AddAction, DeleteAction, RemoveAction, SetAction
    )


_T = TypeVar('_T')
_KT = TypeVar('_KT', bound=str)
_VT = TypeVar('_VT')
_MT = TypeVar('_MT', bound='MapAttribute')
_ACT = TypeVar('_ACT', bound = 'AttributeContainer')

_A = TypeVar('_A', bound='Attribute')

_IMMUTABLE_TYPES = (str, int, float, datetime, timedelta, bytes, bool, tuple, frozenset, type(None))
_IMMUTABLE_TYPE_NAMES = ', '.join(map(lambda x: x.__name__, _IMMUTABLE_TYPES))


class Attribute(Generic[_T]):
    """
    An attribute of a model or an index.

    :param hash_key: If `True`, this attribute is a model's or an index's hash key (partition key).
    :param range_key: If `True`, this attribute is a model's or an index's range key (sort key).
    :param null: If `True`, a `None` value would be considered valid and would result in the attribute
      not being set in the underlying DynamoDB item. If `False` (default), an exception will be raised when
      the attribute is persisted with a `None` value.

      .. note::
         This is different from :class:`pynamodb.attributes.NullAttribute`, which manifests in a `NULL`-typed
         DynamoDB attribute value.

    :param default: A default value that will be assigned in new models (when they are initialized)
      and existing models (when they are loaded).

      .. note::
         Starting with PynamoDB 6.0, the default must be either an immutable value (of one of the built-in
         immutable types) or a callable. This prevents a common class of errors caused by unintentionally mutating
         the default value. A simple workaround is to pass an initializer (e.g. change :code:`default={}` to
         :code:`default=dict`) or wrap in a lambda (e.g. change :code:`default={'foo': 'bar'}` to
         :code:`default=lambda: {'foo': 'bar'}`).

    :param default_for_new: Like `default`, but used only for new models. Use this to assign a default
      for new models that you don't want to apply to existing models when they are loaded and then re-saved.

      .. note::
         Starting with PynamoDB 6.0, the default must be either an immutable value (of one of the built-in
         immutable types) or a callable.

    :param attr_name: The name that is used for the attribute in the underlying DynamoDB item;
        use this to assign a "pythonic" name that is different from the persisted name, i.e.

        .. code-block:: python

          number_of_threads = NumberAttribute(attr_name='thread_count')
    """
    attr_type: str
    null = False

    def __init__(
        self,
        hash_key: bool = False,
        range_key: bool = False,
        null: Optional[bool] = None,
        default: Optional[Union[_T, Callable[..., _T]]] = None,
        default_for_new: Optional[Union[Any, Callable[..., _T]]] = None,
        attr_name: Optional[str] = None,
    ) -> None:
        if default and default_for_new:
            raise ValueError("An attribute cannot have both default and default_for_new parameters")
        if not callable(default) and not isinstance(default, _IMMUTABLE_TYPES):
            raise ValueError(
                f"An attribute's 'default' must be immutable ({_IMMUTABLE_TYPE_NAMES}) or a callable "
                "(see https://pynamodb.readthedocs.io/en/latest/api.html#pynamodb.attributes.Attribute)"
            )
        if not callable(default_for_new) and not isinstance(default_for_new, _IMMUTABLE_TYPES):
            raise ValueError(
                f"An attribute's 'default_for_new' must be immutable ({_IMMUTABLE_TYPE_NAMES}) or a callable "
                "(see https://pynamodb.readthedocs.io/en/latest/api.html#pynamodb.attributes.Attribute)"
            )
        self.default = default
        self.default_for_new = default_for_new

        if null is not None:
            self.null = null
        self.is_hash_key = hash_key
        self.is_range_key = range_key

        # __set_name__ will ensure this is a string
        self.attr_path: List[str] = [attr_name]  # type: ignore

    @property
    def attr_name(self) -> str:
        return self.attr_path[-1]

    @attr_name.setter
    def attr_name(self, value: str) -> None:
        self.attr_path[-1] = value

    def __set__(self, instance: Any, value: Optional[_T]) -> None:
        if instance and not self._is_map_attribute_class_object(instance):
            attr_name = instance._dynamo_to_python_attrs.get(self.attr_name, self.attr_name)
            instance.attribute_values[attr_name] = value

    @overload
    def __get__(self: _A, instance: None, owner: Any) -> _A: ...

    @overload
    def __get__(self: _A, instance: Any, owner: Any) -> _T: ...

    def __get__(self: _A, instance: Any, owner: Any) -> Union[_A, _T]:
        if self._is_map_attribute_class_object(instance):
            # MapAttribute class objects store a local copy of the attribute with `attr_path` set to the document path.
            attr_name = instance._dynamo_to_python_attrs.get(self.attr_name, self.attr_name)
            return instance.__dict__.get(attr_name, None) or self
        elif instance:
            attr_name = instance._dynamo_to_python_attrs.get(self.attr_name, self.attr_name)
            return instance.attribute_values.get(attr_name, None)
        else:
            return self

    def __set_name__(self, owner: Type[Any], name: str) -> None:
        self.attr_name = self.attr_name or name

    def _is_map_attribute_class_object(self, instance: 'Attribute') -> bool:
        return isinstance(instance, MapAttribute) and not instance._is_attribute_container()

    def serialize(self, value: Any) -> Any:
        """
        Serializes a value for botocore's DynamoDB client.

        For a list of DynamoDB attribute types and their matching botocore Python types,
        see `DynamoDB.Client.get_item API reference
        <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb.html#DynamoDB.Client.get_item>`_.
        """
        return value

    def deserialize(self, value: Any) -> Any:
        """
        Deserializes a value from botocore's DynamoDB client.

        For a list of DynamoDB attribute types and their matching botocore Python types,
        see `DynamoDB.Client.get_item API reference
        <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb.html#DynamoDB.Client.get_item>`_.
        """
        return value

    def get_value(self, value: Dict[str, Any]) -> Any:
        if self.attr_type not in value:
            raise AttributeDeserializationError(self.attr_name, self.attr_type)
        return value[self.attr_type]

    def __iter__(self):
        # Because we define __getitem__ below for condition expression support
        raise TypeError("'{}' object is not iterable".format(self.__class__.__name__))

    # Condition Expression Support
    def __eq__(self, other: Any) -> 'Comparison':  # type: ignore[override]
        return Path(self).__eq__(other)

    def __ne__(self, other: Any) -> 'Comparison':  # type: ignore[override]
        return Path(self).__ne__(other)

    def __lt__(self, other: Any) -> 'Comparison':
        return Path(self).__lt__(other)

    def __le__(self, other: Any) -> 'Comparison':
        return Path(self).__le__(other)

    def __gt__(self, other: Any) -> 'Comparison':
        return Path(self).__gt__(other)

    def __ge__(self, other: Any) -> 'Comparison':
        return Path(self).__ge__(other)

    def __getitem__(self, item: Union[int, str]) -> Path:
        return Path(self).__getitem__(item)

    def between(self, lower: Any, upper: Any) -> 'Between':
        return Path(self).between(lower, upper)

    def is_in(self, *values: _T) -> 'In':
        return Path(self).is_in(*values)

    def exists(self) -> 'Exists':
        return Path(self).exists()

    def does_not_exist(self) -> 'NotExists':
        return Path(self).does_not_exist()

    def is_type(self):
        # What makes sense here? Are we using this to check if deserialization will be successful?
        return Path(self).is_type(self.attr_type)

    def startswith(self, prefix: str) -> 'BeginsWith':
        return Path(self).startswith(prefix)

    def contains(self, item: Any) -> 'Contains':
        return Path(self).contains(item)

    # Update Expression Support
    def __add__(self, other: Any) -> '_Increment':
        return Path(self).__add__(other)

    def __radd__(self, other: Any) -> '_Increment':
        return Path(self).__radd__(other)

    def __sub__(self, other: Any) -> '_Decrement':
        return Path(self).__sub__(other)

    def __rsub__(self, other: Any) -> '_Decrement':
        return Path(self).__rsub__(other)

    def __or__(self, other: Any) -> '_IfNotExists':
        return Path(self).__or__(other)

    def append(self, other: Iterable) -> '_ListAppend':
        return Path(self).append(other)

    def prepend(self, other: Iterable) -> '_ListAppend':
        return Path(self).prepend(other)

    def set(
        self,
        value: Union[_T, 'Attribute[_T]', '_Increment', '_Decrement', '_IfNotExists', '_ListAppend']
    ) -> 'SetAction':
        return Path(self).set(value)

    def remove(self) -> 'RemoveAction':
        return Path(self).remove()

    def add(self, *values: Any) -> 'AddAction':
        return Path(self).add(*values)

    def delete(self, *values: Any) -> 'DeleteAction':
        return Path(self).delete(*values)


class AttributeContainerMeta(type):
    _attributes: Dict[str, Attribute]

    def __new__(cls, name, bases, namespace, discriminator=None):
        # Defined so that the discriminator can be set in the class definition.
        return super().__new__(cls, name, bases, namespace)

    def __init__(self, name, bases, namespace, discriminator=None):
        super().__init__(name, bases, namespace)
        AttributeContainerMeta._initialize_attributes(self, discriminator)

    @staticmethod
    def _initialize_attributes(cls, discriminator_value):
        """
        Initialize attributes on the class.
        """
        cls._attributes = {}
        cls._dynamo_to_python_attrs = {}

        for name, attribute in getmembers(cls, lambda o: isinstance(o, Attribute)):
            cls._attributes[name] = attribute
            if attribute.attr_name != name:
                cls._dynamo_to_python_attrs[attribute.attr_name] = name

        # Register the class with the discriminator if necessary.
        discriminators = [name for name, attr in cls._attributes.items() if isinstance(attr, DiscriminatorAttribute)]
        if len(discriminators) > 1:
            raise ValueError("{} has more than one discriminator attribute: {}".format(
                cls.__name__, ", ".join(discriminators)))
        cls._discriminator = discriminators[0] if discriminators else None
        if discriminator_value is not None:
            if not cls._discriminator:
                raise ValueError("{} does not have a discriminator attribute".format(cls.__name__))
            cls._attributes[cls._discriminator].register_class(cls, discriminator_value)


class AttributeContainer(metaclass=AttributeContainerMeta):
    """
    Base class for models and maps.
    """

    def __init__(self, _user_instantiated: bool = True, **attributes: Attribute) -> None:
        # The `attribute_values` dictionary is used by the Attribute data descriptors in cls._attributes
        # to store the values that are bound to this instance. Attributes store values in the dictionary
        # using the `python_attr_name` as the dictionary key. "Raw" (i.e. non-subclassed) MapAttribute
        # instances do not have any Attributes defined and instead use this dictionary to store their
        # collection of name-value pairs.
        self.attribute_values: Dict[str, Any] = {}
        self._set_discriminator()
        self._set_defaults(_user_instantiated=_user_instantiated)
        self._set_attributes(**attributes)

    @classmethod
    def _get_attributes(cls) -> Dict[str, Attribute]:
        """
        Returns the attributes of this class as a mapping from `python_attr_name` => `attribute`.
        """
        warnings.warn("`Model._get_attributes` is deprecated in favor of `Model.get_attributes` now")
        return cls.get_attributes()

    @classmethod
    def get_attributes(cls) -> Dict[str, Attribute]:
        """
        Returns the attributes of this class as a mapping from `python_attr_name` => `attribute`.
        """
        return cls._attributes

    @classmethod
    def _dynamo_to_python_attr(cls, dynamo_key: str) -> str:
        """
        Convert a DynamoDB attribute name to the internal Python name.

        This covers cases where an attribute name has been overridden via "attr_name".
        """
        return cls._dynamo_to_python_attrs.get(dynamo_key, dynamo_key)  # type: ignore

    @classmethod
    def _get_discriminator_attribute(cls) -> Optional['DiscriminatorAttribute']:
        return cls.get_attributes()[cls._discriminator] if cls._discriminator else None  # type: ignore

    def _set_discriminator(self) -> None:
        discriminator_attr = self._get_discriminator_attribute()
        if discriminator_attr and discriminator_attr.get_discriminator(self.__class__) is not None:
            setattr(self, self._discriminator, self.__class__)  # type: ignore

    def _set_defaults(self, _user_instantiated: bool = True) -> None:
        """
        Sets and fields that provide a default value
        """
        for name, attr in self.get_attributes().items():
            if _user_instantiated and attr.default_for_new is not None:
                default = attr.default_for_new
            else:
                default = attr.default
            if callable(default):
                value = default()
            else:
                value = default
            if value is not None:
                setattr(self, name, value)

    def _set_attributes(self, **attributes: Attribute) -> None:
        """
        Sets the attributes for this object
        """
        for attr_name, attr_value in attributes.items():
            if attr_name not in self.get_attributes():
                raise ValueError("Attribute {} specified does not exist".format(attr_name))
            setattr(self, attr_name, attr_value)

    def _container_serialize(self, null_check: bool = True) -> Dict[str, Dict[str, Any]]:
        """
        Serialize attribute values for DynamoDB
        """
        attribute_values: Dict[str, Dict[str, Any]] = {}
        for name, attr in self.get_attributes().items():
            value = getattr(self, name)
            try:
                if isinstance(value, MapAttribute) and not value.validate(null_check=null_check):
                    raise ValueError("Attribute '{}' is not correctly typed".format(name))

                if value is not None:
                    if isinstance(attr, (ListAttribute, MapAttribute)):
                        attr_value = attr.serialize(value, null_check=null_check)
                    else:
                        attr_value = attr.serialize(value)
                else:
                    attr_value = None
            except AttributeNullError as e:
                e.prepend_path(name)
                raise

            if null_check and attr_value is None and not attr.null:
                raise AttributeNullError(name)

            if attr_value is not None:
                attribute_values[attr.attr_name] = {attr.attr_type: attr_value}
        return attribute_values

    def _container_deserialize(self, attribute_values: Dict[str, Dict[str, Any]]) -> None:
        """
        Sets attributes sent back from DynamoDB on this object
        """
        self.attribute_values = {}
        self._set_discriminator()
        self._set_defaults(_user_instantiated=False)
        for name, attr in self.get_attributes().items():
            attribute_value = attribute_values.get(attr.attr_name)
            if attribute_value and NULL not in attribute_value:
                value = attr.deserialize(attr.get_value(attribute_value))
                setattr(self, name, value)

    @classmethod
    def _update_attribute_types(cls, attribute_values: Dict[str, Dict[str, Any]]):
        """
        Update the attribute types in the attribute values dictionary to disambiguate json string and array types
        """
        for attr in cls.get_attributes().values():
            attribute_value = attribute_values.get(attr.attr_name)
            if attribute_value:
                AttributeContainer._coerce_attribute_type(attr.attr_type, attribute_value)
                if isinstance(attr, ListAttribute) and attr.element_type and LIST in attribute_value:
                    if issubclass(attr.element_type, AttributeContainer):
                        for element in attribute_value[LIST]:
                            if MAP in element:
                                attr.element_type._update_attribute_types(element[MAP])
                    else:
                        for element in attribute_value[LIST]:
                            AttributeContainer._coerce_attribute_type(attr.element_type.attr_type, element)
                if isinstance(attr, AttributeContainer) and MAP in attribute_value:
                    attr._update_attribute_types(attribute_value[MAP])

    @staticmethod
    def _coerce_attribute_type(attr_type: str, attribute_value: Dict[str, Any]):
        # coerce attribute types to disambiguate json string and array types
        if attr_type == BINARY and STRING in attribute_value:
            attribute_value[BINARY] = base64.b64decode(attribute_value.pop(STRING))
        elif attr_type == BINARY_SET and LIST in attribute_value:
            attribute_value[BINARY_SET] = [base64.b64decode(v[STRING]) for v in attribute_value.pop(LIST)]
        elif attr_type in {NUMBER_SET, STRING_SET} and LIST in attribute_value:
            json_type = NUMBER if attr_type == NUMBER_SET else STRING
            if all(next(iter(v)) == json_type for v in attribute_value[LIST]):
                attribute_value[attr_type] = [value[json_type] for value in attribute_value.pop(LIST)]

    @classmethod
    def _get_discriminator_class(cls, attribute_values: Dict[str, Dict[str, Any]]) -> Optional[Type]:
        discriminator_attr = cls._get_discriminator_attribute()
        if discriminator_attr:
            discriminator_attribute_value = attribute_values.get(discriminator_attr.attr_name, None)
            if discriminator_attribute_value:
                discriminator_value = discriminator_attr.get_value(discriminator_attribute_value)
                return discriminator_attr.deserialize(discriminator_value)
        return None

    @classmethod
    def _instantiate(cls: Type[_ACT], attribute_values: Dict[str, Dict[str, Any]]) -> _ACT:
        stored_cls = cls._get_discriminator_class(attribute_values)
        if stored_cls and not issubclass(stored_cls, cls):
            raise ValueError("Cannot instantiate a {} from the returned class: {}".format(
                cls.__name__, stored_cls.__name__))
        instance = (stored_cls or cls)(_user_instantiated=False)
        AttributeContainer._container_deserialize(instance, attribute_values)
        return instance

    def to_dynamodb_dict(self) -> Dict[str, Dict[str, Any]]:
        """
        Returns the contents of this instance as a JSON-serializable mapping,
        where each attribute is represented as a mapping with the attribute
        type as the key and the attribute value as the value, e.g.

        .. code-block:: python

           {
               "id": {
                   "N": "12345"
               },
               "name": {
                   "S": "Alice"
               },
           }

        This matches the structure of the "DynamoDB" JSON mapping in the AWS Console.
        """
        attr_values = self._container_serialize(null_check=False)
        for v in attr_values.values():
            bin_encode_attr(v)
        return attr_values

    def from_dynamodb_dict(self, d: Dict[str, Dict[str, Any]]) -> None:
        """
        Sets attributes from a mapping previously produced by :func:`to_dynamodb_dict`.
        """
        for v in d.values():
            bin_decode_attr(v)
        self._update_attribute_types(d)
        self._container_deserialize(d)

    def to_simple_dict(self, *, force: bool = False) -> Dict[str, Any]:
        """
        Returns the contents of this instance as a simple JSON-serializable mapping.

        .. code-block:: python

           {
               "id": 12345,
               "name": "Alice",
           }

        This matches the structure of the "normal" JSON mapping in the AWS Console.

        .. note::

           This representation is limited: by default, it cannot represent binary or set attributes,
           as their encoded form is indistinguishable from a string or list attribute respectively
           (and therefore ambiguous).

        :param force: If :code:`True`, force the conversion even if the model contains Binary or Set attributes
          If :code:`False`, a :code:`ValueError` will be raised if such attributes are set.
        """
        return {k: attr_value_to_simple_dict(v, force) for k, v in self._container_serialize(null_check=False).items()}

    def from_simple_dict(self, d: Dict[str, Any]) -> None:
        """
        Sets attributes from a mapping previously produced by :func:`to_simple_dict`.
        """
        attribute_values = {k: simple_dict_to_attr_value(v) for k, v in d.items()}
        self._update_attribute_types(attribute_values)
        self._container_deserialize(attribute_values)

    def __repr__(self) -> str:
        fields = ', '.join(f'{k}={v!r}' for k, v in self.attribute_values.items())
        return f'{type(self).__name__}({fields})'


class DiscriminatorAttribute(Attribute[type]):
    attr_type = STRING

    def __init__(self, attr_name: Optional[str] = None) -> None:
        super().__init__(attr_name=attr_name)
        self._class_map: Dict[type, Any] = {}
        self._discriminator_map: Dict[Any, type] = {}

    def register_class(self, cls: type, discriminator: Any):
        discriminator = discriminator(cls) if callable(discriminator) else discriminator
        current_class = self._discriminator_map.get(discriminator)
        if current_class and current_class != cls:
            raise ValueError("The discriminator value '{}' is already assigned to a class: {}".format(
                discriminator, current_class.__name__))

        if cls not in self._class_map:
            self._class_map[cls] = discriminator

        self._discriminator_map[discriminator] = cls

    def get_registered_subclasses(self, cls: Type[_T]) -> List[Type[_T]]:
        return [k for k in self._class_map.keys() if issubclass(k, cls)]

    def get_discriminator(self, cls: type) -> Optional[Any]:
        return self._class_map.get(cls)

    def __set__(self, instance: Any, value: Optional[type]) -> None:
        if type(instance) != value:
            raise ValueError("The discriminator attribute must be set to the instance type: {}".format(type(instance)))
        super().__set__(instance, value)

    def serialize(self, value):
        """
        Returns the discriminator value corresponding to the given class.
        """
        return self._class_map[value]

    def deserialize(self, value):
        """
        Returns the class corresponding to the given discriminator value.
        """
        if value not in self._discriminator_map:
            raise ValueError("Unknown discriminator value: {}".format(value))
        return self._discriminator_map[value]


class BinaryAttribute(Attribute[bytes]):
    """
    An attribute containing a binary data object (:code:`bytes`).

    :param legacy_encoding: If :code:`True`, inefficient legacy encoding will be used to maintain compatibility
      with PynamoDB 5 and lower. Set to :code:`False` for new tables and models, and always set to :code:`False`
      within :class:`~pynamodb.attributes.MapAttribute`.

      For more details, see :doc:`upgrading_binary`.
    """
    attr_type = BINARY

    def __init__(self, *args: Any, legacy_encoding: bool, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.legacy_encoding = legacy_encoding

    def serialize(self, value):
        if self.legacy_encoding:
            return b64encode(value)
        return value

    def deserialize(self, value):
        if self.legacy_encoding:
            return b64decode(value)
        return value


class BinarySetAttribute(Attribute[Set[bytes]]):
    """
    An attribute containing a set of binary data objects (:code:`bytes`).

    :param legacy_encoding: If :code:`True`, inefficient legacy encoding will be used to maintain compatibility
      with PynamoDB 5 and lower. Set to :code:`False` for new tables and models, and always set to :code:`False`
      within :class:`~pynamodb.attributes.MapAttribute`.

      For more details, see :doc:`upgrading_binary`.
    """
    attr_type = BINARY_SET
    null = True

    def __init__(self, *args: Any, legacy_encoding: bool, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.legacy_encoding = legacy_encoding

    def serialize(self, value):
        """
        Returns a list of base64 encoded binary strings. Encodes empty sets as "None".
        """
        if self.legacy_encoding:
            return [b64encode(v) for v in value] or None
        return list(value) or None

    def deserialize(self, value):
        """
        Returns a set of decoded byte strings from base64 encoded values.
        """
        if self.legacy_encoding:
            return {b64decode(v) for v in value}
        return set(value)


class UnicodeAttribute(Attribute[str]):
    """
    A unicode attribute
    """
    attr_type = STRING


class UnicodeSetAttribute(Attribute[Set[str]]):
    """
    A unicode set
    """
    attr_type = STRING_SET
    null = True

    def serialize(self, value):
        """
        Returns a list of strings. Encodes empty sets as "None".
        """
        return list(value) or None

    def deserialize(self, value):
        """
        Returns a set from a list of strings.
        """
        return set(value)


class JSONAttribute(Attribute[Any]):
    """
    A JSON Attribute

    Encodes JSON to unicode internally
    """
    attr_type = STRING

    def serialize(self, value) -> Optional[str]:
        """
        Serializes JSON to unicode
        """
        if value is None:
            return None
        encoded = json.dumps(value)
        return encoded

    def deserialize(self, value):
        """
        Deserializes JSON
        """
        return json.loads(value, strict=False)


class BooleanAttribute(Attribute[bool]):
    """
    A class for boolean attributes
    """
    attr_type = BOOLEAN

    def serialize(self, value):
        if value is None:
            return None
        elif value:
            return True
        else:
            return False

    def deserialize(self, value):
        return bool(value)


class NumberAttribute(Attribute[float]):
    """
    A number attribute
    """
    attr_type = NUMBER

    def serialize(self, value):
        """
        Encode numbers as JSON
        """
        return json.dumps(value)

    def deserialize(self, value):
        """
        Decode numbers from JSON
        """
        return json.loads(value)


class NumberSetAttribute(Attribute[Set[float]]):
    """
    A number set attribute
    """
    attr_type = NUMBER_SET
    null = True

    def serialize(self, value):
        """
        Encodes a set of numbers as a JSON list. Encodes empty sets as "None".
        """
        return [json.dumps(v) for v in value] or None

    def deserialize(self, value):
        """
        Returns a set from a JSON list of numbers.
        """
        return {json.loads(v) for v in value}


class VersionAttribute(NumberAttribute):
    """
    A version attribute
    """
    null = True

    def __set__(self, instance, value):
        """
        Cast assigned value to int.
        """
        super().__set__(instance, int(value))

    def __get__(self, instance, owner):
        """
        Cast retrieved value to int.
        """
        val = super().__get__(instance, owner)
        return int(val) if isinstance(val, float) else val

    def serialize(self, value):
        """
        Cast value to int then encode as JSON
        """
        return super().serialize(int(value))

    def deserialize(self, value):
        """
        Decode numbers from JSON and cast to int.
        """
        return int(super().deserialize(value))


class TTLAttribute(Attribute[datetime]):
    """
    A time-to-live attribute that signifies when the item expires and can be automatically deleted.
    It can be assigned with a timezone-aware datetime value (for absolute expiry time)
    or a timedelta value (for expiry relative to the current time),
    but always reads as a UTC datetime value.
    """
    attr_type = NUMBER

    def _normalize(self, value):
        """
        Converts value to a UTC datetime
        """
        if value is None:
            return
        if isinstance(value, timedelta):
            value = int(time.time() + value.total_seconds())
        elif isinstance(value, datetime):
            if value.tzinfo is None:
                raise ValueError("datetime must be timezone-aware")
            value = calendar.timegm(value.utctimetuple())
        else:
            raise ValueError("TTLAttribute value must be a timedelta or datetime")
        return datetime.utcfromtimestamp(value).replace(tzinfo=timezone.utc)

    def __set__(self, instance, value):
        """
        Converts assigned values to a UTC datetime
        """
        super().__set__(instance, self._normalize(value))

    def serialize(self, value):
        """
        Serializes a datetime as a timestamp (Unix time).
        """
        if value is None:
            return None
        return json.dumps(calendar.timegm(self._normalize(value).utctimetuple()))

    def deserialize(self, value):
        """
        Deserializes a timestamp (Unix time) as a UTC datetime.
        """
        timestamp = json.loads(value)
        return datetime.utcfromtimestamp(timestamp).replace(tzinfo=timezone.utc)


class UTCDateTimeAttribute(Attribute[datetime]):
    """
    An attribute for storing a UTC Datetime
    """
    attr_type = STRING

    def serialize(self, value):
        """
        Takes a datetime object and returns a string
        """
        if value.tzinfo is None:
            value = value.replace(tzinfo=timezone.utc)
        # Padding of years under 1000 is inconsistent and depends on system strftime:
        # https://bugs.python.org/issue13305
        fmt = value.astimezone(timezone.utc).strftime(DATETIME_FORMAT).zfill(31)
        return fmt

    def deserialize(self, value):
        """
        Takes a UTC datetime string and returns a datetime object
        """
        return self._fast_parse_utc_date_string(value)

    @staticmethod
    def _fast_parse_utc_date_string(date_string: str) -> datetime:
        # Method to quickly parse strings formatted with '%Y-%m-%dT%H:%M:%S.%f+0000'.
        # This is ~5.8x faster than using strptime and 38x faster than dateutil.parser.parse.
        _int = int  # Hack to prevent global lookups of int, speeds up the function ~10%
        try:
            # Fix pre-1000 dates serialized on systems where strftime doesn't pad w/older PynamoDB versions.
            date_string = date_string.zfill(31)
            if (len(date_string) != 31 or date_string[4] != '-' or date_string[7] != '-'
                    or date_string[10] != 'T' or date_string[13] != ':' or date_string[16] != ':'
                    or date_string[19] != '.' or date_string[26:31] != '+0000'):
                raise ValueError("Datetime string '{}' does not match format '{}'".format(date_string, DATETIME_FORMAT))
            return datetime(
                _int(date_string[0:4]), _int(date_string[5:7]), _int(date_string[8:10]),
                _int(date_string[11:13]), _int(date_string[14:16]), _int(date_string[17:19]),
                _int(date_string[20:26]), timezone.utc
            )
        except (TypeError, ValueError):
            raise ValueError("Datetime string '{}' does not match format '{}'".format(date_string, DATETIME_FORMAT))


class NullAttribute(Attribute[None]):
    attr_type = NULL

    def serialize(self, value):
        return True

    def deserialize(self, value):
        return None


class MetaMapAttribute(AttributeContainerMeta):
    def __init__(self, name, bases, namespace, discriminator=None):
        super().__init__(name, bases, namespace, discriminator=discriminator)
        for attr_name, attr in self._attributes.items():
            if isinstance(attr, (BinaryAttribute, BinarySetAttribute)) and attr.legacy_encoding:
                raise ValueError(
                    "Legacy encoding is only ever needed for top-level (model) attributes. "
                    f"Please remove the legacy_encoding flag from the definition of '{attr_name}'."
                )


class MapAttribute(Attribute[Mapping[_KT, _VT]], AttributeContainer, metaclass=MetaMapAttribute):
    """
    A Map Attribute

    The MapAttribute class can be used to store a JSON document as "raw" name-value pairs, or
    it can be subclassed and the document fields represented as class attributes using Attribute instances.

    To support the ability to subclass MapAttribute and use it as an AttributeContainer, instances of
    MapAttribute behave differently based both on where they are instantiated and on their type.
    Because of this complicated behavior, a bit of an introduction is warranted.

    Models that contain a MapAttribute define its properties using a class attribute on the model.
    For example, below we define "MyModel" which contains a MapAttribute "my_map":

    class MyModel(Model):
       my_map = MapAttribute(attr_name="dynamo_name", default={})

    When instantiated in this manner (as a class attribute of an AttributeContainer class), the MapAttribute
    class acts as an instance of the Attribute class. The instance stores data about the attribute (in this
    example the dynamo name and default value), and acts as a data descriptor, storing any value bound to it
    on the `attribute_values` dictionary of the containing instance (in this case an instance of MyModel).

    Unlike other Attribute types, the value that gets bound to the containing instance is a new instance of
    MapAttribute, not an instance of the primitive type. For example, a UnicodeAttribute stores strings in
    the `attribute_values` of the containing instance; a MapAttribute does not store a dict but instead stores
    a new instance of itself. This difference in behavior is necessary when subclassing MapAttribute in order
    to access the Attribute data descriptors that represent the document fields.

    For example, below we redefine "MyModel" to use a subclass of MapAttribute as "my_map":

    class MyMapAttribute(MapAttribute):
        my_internal_map = MapAttribute()

    class MyModel(Model):
        my_map = MyMapAttribute(attr_name="dynamo_name", default = {})

    In order to set the value of my_internal_map on an instance of MyModel we need the bound value for "my_map"
    to be an instance of MapAttribute so that it acts as a data descriptor:

    MyModel().my_map.my_internal_map = {'foo': 'bar'}

    That is the attribute access of "my_map" must return a MyMapAttribute instance and not a dict.

    When an instance is used in this manner (bound to an instance of an AttributeContainer class),
    the MapAttribute class acts as an AttributeContainer class itself. The instance does not store data
    about the attribute, and does not act as a data descriptor. The instance stores name-value pairs in its
    internal `attribute_values` dictionary.

    Thus while MapAttribute multiply inherits from Attribute and AttributeContainer, a MapAttribute instance
    does not behave as both an Attribute AND an AttributeContainer. Rather an instance of MapAttribute behaves
    EITHER as an Attribute OR as an AttributeContainer, depending on where it was instantiated.

    So, how do we create this dichotomous behavior?
    All MapAttribute instances are initialized as AttributeContainers only. During construction of
    AttributeContainer classes (subclasses of MapAttribute and Model), any instances that are class attributes
    are transformed from AttributeContainers to Attributes (via the `_make_attribute` method call).
    """
    attr_type = MAP

    attribute_args = getfullargspec(Attribute.__init__).args[1:]

    def __init__(self, **attributes):
        # Store the kwargs used by Attribute.__init__ in case `_make_attribute` is called.
        self.attribute_kwargs = {arg: attributes.pop(arg) for arg in self.attribute_args if arg in attributes}

        # Assume all instances should behave like an AttributeContainer. Instances that are intended to be
        # used as Attributes will be transformed during creation of the containing class.
        # Because of this do not use MRO or cooperative multiple inheritance, call the parent class directly.
        AttributeContainer.__init__(self, **attributes)

        # It is possible that attributes names can collide with argument names of Attribute.__init__.
        # Assume that this is the case if any of the following are true:
        #   - the user passed in other attributes that did not match any argument names
        #   - this is a "raw" (i.e. non-subclassed) MapAttribute instance and attempting to store the attributes
        #     cannot raise a ValueError (if this assumption is wrong, calling `_make_attribute` removes them)
        #   - the names of all attributes in self.attribute_kwargs match attributes defined on the class
        if self.attribute_kwargs and (
                attributes or self.is_raw() or all(arg in self.get_attributes() for arg in self.attribute_kwargs)):
            self._set_attributes(**self.attribute_kwargs)

    def _is_attribute_container(self):
        # Determine if this instance is being used as an AttributeContainer or an Attribute.
        # AttributeContainer instances have an internal `attribute_values` dictionary that is removed
        # by the `_make_attribute` call during initialization of the containing class.
        return 'attribute_values' in self.__dict__

    def _make_attribute(self):
        # WARNING! This function is only intended to be called from the __set_name__ function.
        if not self._is_attribute_container():
            raise AssertionError("MapAttribute._make_attribute called on an initialized instance")
        # During initialization the kwargs were stored in `attribute_kwargs`. Remove them and re-initialize the class.
        kwargs = self.attribute_kwargs
        del self.attribute_kwargs
        del self.attribute_values
        Attribute.__init__(self, **kwargs)
        for name, attr in self.get_attributes().items():
            # Set a local attribute with the same name that shadows the class attribute.
            # Because attr is a data descriptor and the attribute already exists on the class,
            # we have to store the local copy directly into __dict__ to prevent calling attr.__set__.
            # Use deepcopy so that `attr_path` and any local attributes are also copied.
            self.__dict__[name] = deepcopy(attr)

    def _update_attribute_paths(self, path_segment):
        # WARNING! This function is only intended to be called from the __set_name__ function.
        if self._is_attribute_container():
            raise AssertionError("MapAttribute._update_attribute_paths called before MapAttribute._make_attribute")
        for name in self.get_attributes().keys():
            local_attr = self.__dict__[name]
            local_attr.attr_path.insert(0, path_segment)
            if isinstance(local_attr, MapAttribute):
                local_attr._update_attribute_paths(path_segment)

    def __eq__(self, other: Any) -> 'Comparison':  # type: ignore[override]
        if self._is_attribute_container():
            return NotImplemented
        return Attribute.__eq__(self, other)

    def __ne__(self, other: Any) -> 'Comparison':  # type: ignore[override]
        if self._is_attribute_container():
            return NotImplemented
        return Attribute.__ne__(self, other)

    def __lt__(self, other: Any) -> 'Comparison':
        if self._is_attribute_container():
            return NotImplemented
        return Attribute.__lt__(self, other)

    def __le__(self, other: Any) -> 'Comparison':
        if self._is_attribute_container():
            return NotImplemented
        return Attribute.__le__(self, other)

    def __gt__(self, other: Any) -> 'Comparison':
        if self._is_attribute_container():
            return NotImplemented
        return Attribute.__gt__(self, other)

    def __ge__(self, other: Any) -> 'Comparison':
        if self._is_attribute_container():
            return NotImplemented
        return Attribute.__ge__(self, other)

    def __iter__(self):
        if self._is_attribute_container():
            return iter(self.attribute_values)
        return super().__iter__()

    def __getitem__(self, item: _KT) -> _VT:  # type: ignore
        if self._is_attribute_container():
            return self.attribute_values[item]
        # If this instance is being used as an Attribute, treat item access like the map dereference operator.
        # This provides equivalence between DynamoDB's nested attribute access for map elements (MyMap.nestedField)
        # and Python's item access for dictionaries (MyMap['nestedField']).
        if item in self.get_attributes():
            return getattr(self, item)
        elif self.is_raw():
            return Path(self.attr_path + [str(item)])  # type: ignore
        else:
            raise AttributeError("'{}' has no attribute '{}'".format(self.__class__.__name__, item))

    def __setitem__(self, item, value):
        if not self._is_attribute_container():
            raise TypeError("'{}' object does not support item assignment".format(self.__class__.__name__))
        if item in self.get_attributes():
            setattr(self, item, value)
        elif self.is_raw():
            self.attribute_values[item] = value
        else:
            raise AttributeError("'{}' has no attribute '{}'".format(self.__class__.__name__, item))

    def __getattr__(self, attr: str) -> _VT:
        # This should only be called for "raw" (i.e. non-subclassed) MapAttribute instances.
        # MapAttribute subclasses should access attributes via the Attribute descriptors.
        if self.is_raw() and self._is_attribute_container():
            try:
                return self.attribute_values[attr]
            except KeyError:
                pass
        raise AttributeError("'{}' has no attribute '{}'".format(self.__class__.__name__, attr))

    @overload  # type: ignore
    def __get__(self: _A, instance: None, owner: Any) -> _A: ...
    @overload
    def __get__(self: _MT, instance: Any, owner: Any) -> _MT: ...
    def __get__(self: _A, instance: Any, owner: Any) -> Union[_A, _T]:
        # just for typing
        return super().__get__(instance, owner)

    def __setattr__(self, name, value):
        # "Raw" (i.e. non-subclassed) instances set their name-value pairs in the `attribute_values` dictionary.
        # MapAttribute subclasses should set attributes via the Attribute descriptors.
        if self.is_raw() and self._is_attribute_container():
            self.attribute_values[name] = value
        else:
            object.__setattr__(self, name, value)

    def __set__(self, instance: Any, value: Union[None, 'MapAttribute[_KT, _VT]', Mapping[_KT, _VT]]):
        if isinstance(value, collections.abc.Mapping):
            value = type(self)(**value)  # type: ignore
        return super().__set__(instance, value)  # type: ignore

    def __set_name__(self, owner: Type[Any], name: str) -> None:
        if issubclass(owner, AttributeContainer):
            # MapAttribute instances that are class attributes of an AttributeContainer class
            # should behave like an Attribute instance and not an AttributeContainer instance.
            self._make_attribute()

            super().__set_name__(owner, name)

            # To support creating expressions from nested attributes, MapAttribute instances
            # store local copies of the attributes in cls._attributes with `attr_path` set.
            # Prepend the `attr_path` lists with the dynamo attribute name.
            self._update_attribute_paths(self.attr_name)

    def _set_attributes(self, **attrs):
        """
        Sets the attributes for this object
        """
        if self.is_raw():
            for name, value in attrs.items():
                setattr(self, name, value)
        else:
            super()._set_attributes(**attrs)

    def is_correctly_typed(self, key, attr, *, null_check: bool = True):
        can_be_null = attr.null or not null_check
        value = getattr(self, key)
        if can_be_null and value is None:
            return True
        if getattr(self, key) is None:
            raise AttributeNullError(key)
        return True  # TODO: check that the actual type of `value` meets requirements of `attr`

    def validate(self, *, null_check: bool = False):
        return all(self.is_correctly_typed(k, v, null_check=null_check)
                   for k, v in self.get_attributes().items())

    def _serialize_undeclared_attributes(self, values, container: Dict):
        # Continue to serialize NULL values in "raw" map attributes for backwards compatibility.
        # This special case behavior for "raw" attributes should be removed in the future.
        for attr_name in values:
            if attr_name not in self.get_attributes():
                v = values[attr_name]
                attr_class = _get_class_for_serialize(v)
                attr_type = attr_class.attr_type
                attr_value = attr_class.serialize(v)
                if attr_value is None:
                    # When attribute values serialize to "None" (e.g. empty sets) we store {"NULL": True} in DynamoDB.
                    attr_type = NULL
                    attr_value = True
                container[attr_name] = {attr_type: attr_value}
        return container

    def serialize(self, values, *, null_check: bool = True):
        if not self.is_raw():
            # This is a subclassed MapAttribute that acts as an AttributeContainer.
            # Serialize the values based on the attributes in the class.

            if not isinstance(values, type(self)):
                # Copy the values onto an instance of the class for serialization.
                instance = type(self)()
                instance.attribute_values = {}  # clear any defaults
                for name in values:
                    if name in self.get_attributes():
                        setattr(instance, name, values[name])
                values = instance

            return AttributeContainer._container_serialize(values, null_check=null_check)

        # For a "raw" MapAttribute all fields are undeclared
        return self._serialize_undeclared_attributes(values, {})

    def deserialize(self, values):
        """
        Decode as a dict.
        """
        if not self.is_raw():
            # If this is a subclass of a MapAttribute (i.e typed), instantiate an instance
            return self._instantiate(values)

        return {
            k: DESERIALIZE_CLASS_MAP[attr_type].deserialize(attr_value)
            for k, v in values.items() for attr_type, attr_value in v.items()
        }

    @classmethod
    def is_raw(cls):
        return cls == MapAttribute

    def as_dict(self):
        result = {}
        for key, value in self.attribute_values.items():
            result[key] = value.as_dict() if isinstance(value, MapAttribute) else value
        return result


class DynamicMapAttribute(MapAttribute):
    """
    A map attribute that supports declaring attributes (like an AttributeContainer) but will also store
    any other values that are set on it (like a raw MapAttribute).

    >>> class MyDynamicMapAttribute(DynamicMapAttribute):
    >>>     a_date_time = UTCDateTimeAttribute()  # raw map attributes cannot serialize/deserialize datetime values
    >>>
    >>> dynamic_map = MyDynamicMapAttribute()
    >>> dynamic_map.a_date_time = datetime.utcnow()
    >>> dynamic_map.a_number = 5
    >>> dynamic_map.serialize()  # {'a_date_time': {'S': 'xxx'}, 'a_number': {'N': '5'}}
    """

    def __setattr__(self, name, value):
        # Set attributes via the Attribute descriptor if it exists.
        if name in self.get_attributes():
            object.__setattr__(self, name, value)
        else:
            super().__setattr__(name, value)

    def serialize(self, values, *, null_check: bool = True):
        if not isinstance(values, type(self)):
            # Copy the values onto an instance of the class for serialization.
            instance = type(self)()
            instance.attribute_values = {}  # clear any defaults
            instance._set_attributes(**values)
            values = instance

        # this serializes the class defined attributes.
        # we do this first because we have type checks that validate the data
        rval = AttributeContainer._container_serialize(values, null_check=null_check)

        # this serializes the dynamically defined attributes
        # we have no real type safety here so we have to dynamically construct the type to write to dynamo
        self._serialize_undeclared_attributes(values, rval)

        return rval

    def deserialize(self, values):
        # this deserializes the class defined attributes
        # we do this first so that the we populate the defined object attributes fields properly with type safety
        instance = self._instantiate(values)
        # this deserializes the dynamically defined attributes
        for attr_name, value in values.items():
            if instance._dynamo_to_python_attr(attr_name) not in instance.get_attributes():
                attr_type, attr_value = next(iter(value.items()))
                instance[attr_name] = DESERIALIZE_CLASS_MAP[attr_type].deserialize(attr_value)
        return instance

    @classmethod
    def is_raw(cls):
        # All subclasses of DynamicMapAttribute should be treated like "raw" map attributes.
        return True


def _get_class_for_serialize(value: Any) -> Attribute:
    if value is None:
        return NullAttribute()
    if isinstance(value, MapAttribute):
        return value
    if isinstance(value, set):
        set_types = {type(v) for v in value}
        if not set_types:
            raise ValueError("Cannot serialize empty set")
        if set_types == {str}:
            return UnicodeSetAttribute()
        if set_types <= {int, float}:
            return NumberSetAttribute()
        if set_types == {bytes}:
            return BinarySetAttribute(legacy_encoding=False)
        raise ValueError(f"Cannot serialize set consisting of types: {', '.join(sorted(map(repr, set_types)))}")

    value_type = type(value)
    attr = SERIALIZE_CLASS_MAP.get(value_type)
    if attr is None:
        raise ValueError(f"Unsupported value type '{value_type}'")
    return attr


class ListAttribute(Generic[_T], Attribute[List[_T]]):
    attr_type = LIST
    element_type: Optional[Type[Attribute]] = None

    def __init__(
        self,
        hash_key: bool = False,
        range_key: bool = False,
        null: Optional[bool] = None,
        default: Optional[Union[Any, Callable[..., Any]]] = None,
        attr_name: Optional[str] = None,
        of: Optional[Type[_T]] = None,
    ) -> None:
        super().__init__(
            hash_key=hash_key,
            range_key=range_key,
            null=null,
            default=default,
            attr_name=attr_name,
        )
        if of:
            if not issubclass(of, Attribute):
                raise ValueError("'of' must be a subclass of Attribute")
            self.element_type = of

    def serialize(self, values, *, null_check: bool = True):
        """
        Encode the given list of objects into a list of AttributeValue types.
        """
        rval = []
        for idx, value in enumerate(values):
            attr = self._get_serialize_class(value)
            if self.element_type and value is not None and not isinstance(attr, self.element_type):
                raise ValueError("List elements must be of type: {}".format(self.element_type.__name__))
            attr_type = attr.attr_type
            try:
                if isinstance(attr, (ListAttribute, MapAttribute)):
                    attr_value = attr.serialize(value, null_check=null_check)
                else:
                    attr_value = attr.serialize(value)
            except AttributeNullError as e:
                e.prepend_path(f'[{idx}]')
                raise
            if attr_value is None:
                # When attribute values serialize to "None" (e.g. empty sets) we store {"NULL": True} in DynamoDB.
                attr_type = NULL
                attr_value = True
            rval.append({attr_type: attr_value})
        return rval

    def deserialize(self, values):
        """
        Decode from list of AttributeValue types.
        """
        if self.element_type:
            element_attr: Attribute
            if issubclass(self.element_type, (BinaryAttribute, BinarySetAttribute)):
                element_attr = self.element_type(legacy_encoding=False)
            else:
                element_attr = self.element_type()
                if isinstance(element_attr, MapAttribute):
                    element_attr._make_attribute()  # ensure attr_name exists
            deserialized_lst = []
            for idx, attribute_value in enumerate(values):
                value = None
                if NULL not in attribute_value:
                    # set attr_name in case `get_value` raises an exception
                    element_attr.attr_name = f'{self.attr_name}[{idx}]' if self.attr_name else f'[{idx}]'
                    value = element_attr.deserialize(element_attr.get_value(attribute_value))
                deserialized_lst.append(value)
            return deserialized_lst

        return [
            DESERIALIZE_CLASS_MAP[attr_type].deserialize(attr_value)
            for v in values for attr_type, attr_value in v.items()
        ]

    def __getitem__(self, idx: int) -> Path:  # type: ignore
        if not isinstance(idx, int):
            raise TypeError("list indices must be integers, not {}".format(type(idx).__name__))

        if self.element_type:
            # If this instance is typed, return a properly configured attribute on list element access.
            element_attr = self.element_type()
            if isinstance(element_attr, MapAttribute):
                element_attr._make_attribute()
            element_attr.attr_path = list(self.attr_path)  # copy the document path before indexing last element
            element_attr.attr_name = '{}[{}]'.format(element_attr.attr_name, idx)
            if isinstance(element_attr, MapAttribute):
                for path_segment in reversed(element_attr.attr_path):
                    element_attr._update_attribute_paths(path_segment)
            return element_attr  # type: ignore

        return super().__getitem__(idx)

    def _get_serialize_class(self, value):
        if value is None:
            return NullAttribute()
        if isinstance(value, Attribute):
            return value
        if self.element_type:
            if issubclass(self.element_type, (BinaryAttribute, BinarySetAttribute)):
                return self.element_type(legacy_encoding=False)
            return self.element_type()
        return _get_class_for_serialize(value)


DESERIALIZE_CLASS_MAP: Dict[str, Attribute] = {
    BINARY: BinaryAttribute(legacy_encoding=False),
    BINARY_SET: BinarySetAttribute(legacy_encoding=False),
    BOOLEAN: BooleanAttribute(),
    LIST: ListAttribute(),
    MAP: MapAttribute(),
    NULL: NullAttribute(),
    NUMBER: NumberAttribute(),
    NUMBER_SET: NumberSetAttribute(),
    STRING: UnicodeAttribute(),
    STRING_SET: UnicodeSetAttribute()
}

SERIALIZE_CLASS_MAP: Mapping[type, Attribute] = {
    dict: MapAttribute(),
    list: ListAttribute(),
    bool: BooleanAttribute(),
    float: NumberAttribute(),
    int: NumberAttribute(),
    str: UnicodeAttribute(),
    bytes: BinaryAttribute(legacy_encoding=False),
}



import sys
from unittest.mock import patch

import pytest
from importlib import reload

import pynamodb.settings


@pytest.mark.parametrize('settings_str', [
    "session_cls = object()",
    "request_timeout_seconds = 5",
])
def test_override_old_attributes(settings_str, tmpdir):
    custom_settings = tmpdir.join("pynamodb_settings.py")
    custom_settings.write(settings_str)

    with patch.dict('os.environ', {'PYNAMODB_CONFIG': str(custom_settings)}):
        with pytest.warns(UserWarning) as warns:
            reload(pynamodb.settings)
    assert len(warns) == 1
    assert 'options are no longer supported' in str(warns[0].message)



"""
PynamoDB Indexes
"""
from inspect import getmembers
from typing import Any, Dict, Generic, List, Optional, Type, TypeVar
from typing import TYPE_CHECKING

from pynamodb._schema import IndexSchema, GlobalSecondaryIndexSchema
from pynamodb._schema import ModelSchema
from pynamodb.constants import (
    INCLUDE, ALL, KEYS_ONLY, ATTR_NAME, ATTR_TYPE, KEY_TYPE,
    PROJECTION_TYPE, NON_KEY_ATTRIBUTES,
    READ_CAPACITY_UNITS, WRITE_CAPACITY_UNITS,
)
from pynamodb.attributes import Attribute
from pynamodb.expressions.condition import Condition
from pynamodb.pagination import ResultIterator
from pynamodb.types import HASH, RANGE
if TYPE_CHECKING:
    from pynamodb.models import Model

_KeyType = Any
_M = TypeVar('_M', bound='Model')


class Index(Generic[_M]):
    """
    Base class for secondary indexes
    """
    Meta: Any = None
    _model: _M

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls.Meta is not None:
            cls.Meta.attributes = {}
            for name, attribute in getmembers(cls, lambda o: isinstance(o, Attribute)):
                cls.Meta.attributes[name] = attribute

    def __init__(self) -> None:
        if self.Meta is None:
            raise ValueError("Indexes require a Meta class for settings")
        if not hasattr(self.Meta, "projection"):
            raise ValueError("No projection defined, define a projection for this class")

    def __set_name__(self, owner: Type[_M], name: str):
        if not hasattr(self.Meta, "index_name"):
            self.Meta.index_name = name

    def count(
        self,
        hash_key: _KeyType,
        range_key_condition: Optional[Condition] = None,
        filter_condition: Optional[Condition] = None,
        consistent_read: bool = False,
        limit: Optional[int] = None,
        rate_limit: Optional[float] = None,
    ) -> int:
        """
        Count on an index
        """
        return self._model.count(
            hash_key,
            range_key_condition=range_key_condition,
            filter_condition=filter_condition,
            index_name=self.Meta.index_name,
            consistent_read=consistent_read,
            limit=limit,
            rate_limit=rate_limit,
        )

    def query(
        self,
        hash_key: _KeyType,
        range_key_condition: Optional[Condition] = None,
        filter_condition: Optional[Condition] = None,
        consistent_read: bool = False,
        scan_index_forward: Optional[bool] = None,
        limit: Optional[int] = None,
        last_evaluated_key: Optional[Dict[str, Dict[str, Any]]] = None,
        attributes_to_get: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        rate_limit: Optional[float] = None,
    ) -> ResultIterator[_M]:
        """
        Queries an index
        """
        return self._model.query(
            hash_key,
            range_key_condition=range_key_condition,
            filter_condition=filter_condition,
            consistent_read=consistent_read,
            index_name=self.Meta.index_name,
            scan_index_forward=scan_index_forward,
            limit=limit,
            last_evaluated_key=last_evaluated_key,
            attributes_to_get=attributes_to_get,
            page_size=page_size,
            rate_limit=rate_limit,
        )

    def scan(
        self,
        filter_condition: Optional[Condition] = None,
        segment: Optional[int] = None,
        total_segments: Optional[int] = None,
        limit: Optional[int] = None,
        last_evaluated_key: Optional[Dict[str, Dict[str, Any]]] = None,
        page_size: Optional[int] = None,
        consistent_read: Optional[bool] = None,
        rate_limit: Optional[float] = None,
        attributes_to_get: Optional[List[str]] = None,
    ) -> ResultIterator[_M]:
        """
        Scans an index
        """
        return self._model.scan(
            filter_condition=filter_condition,
            segment=segment,
            total_segments=total_segments,
            limit=limit,
            last_evaluated_key=last_evaluated_key,
            page_size=page_size,
            consistent_read=consistent_read,
            index_name=self.Meta.index_name,
            rate_limit=rate_limit,
            attributes_to_get=attributes_to_get,
        )

    @classmethod
    def _hash_key_attribute(cls):
        """
        Returns the attribute class for the hash key
        """
        for attr_cls in cls.Meta.attributes.values():
            if attr_cls.is_hash_key:
                return attr_cls

    def _update_model_schema(self, schema: ModelSchema) -> None:
        raise NotImplementedError

    @classmethod
    def _get_schema(cls) -> IndexSchema:
        """
        Returns the schema for this index
        """
        schema: IndexSchema = {
            'index_name': cls.Meta.index_name,
            'key_schema': [],
            'projection': {
                PROJECTION_TYPE: cls.Meta.projection.projection_type,
            },
            'attribute_definitions': [],
        }

        for attr_cls in cls.Meta.attributes.values():
            if attr_cls.is_hash_key or attr_cls.is_range_key:
                schema['attribute_definitions'].append({
                    ATTR_NAME: attr_cls.attr_name,
                    ATTR_TYPE: attr_cls.attr_type,
                })
                schema['key_schema'].append({
                    ATTR_NAME: attr_cls.attr_name,
                    KEY_TYPE: HASH if attr_cls.is_hash_key else RANGE,
                })
        if cls.Meta.projection.non_key_attributes:
            schema['projection'][NON_KEY_ATTRIBUTES] = cls.Meta.projection.non_key_attributes
        return schema


class GlobalSecondaryIndex(Index[_M]):
    """
    A global secondary index
    """
    @classmethod
    def _update_model_schema(cls, schema: ModelSchema) -> None:
        index_schema: GlobalSecondaryIndexSchema = {
            **cls._get_schema(),  # type:ignore[misc]  # https://github.com/python/mypy/pull/13353
            'provisioned_throughput': {},
        }

        if hasattr(cls.Meta, 'read_capacity_units'):
            index_schema['provisioned_throughput'][READ_CAPACITY_UNITS] = cls.Meta.read_capacity_units
        if hasattr(cls.Meta, 'write_capacity_units'):
            index_schema['provisioned_throughput'][WRITE_CAPACITY_UNITS] = cls.Meta.write_capacity_units

        schema['global_secondary_indexes'].append(index_schema)
        # With polymorphism, indexes can use the same attribute, e.g. index1 on (thread_id, created_at)
        # and index2 on (thread_id, updated_at). We need to deduplicate.
        for attr_def in index_schema['attribute_definitions']:
            if attr_def not in schema['attribute_definitions']:
                schema['attribute_definitions'].append(attr_def)


class LocalSecondaryIndex(Index[_M]):
    """
    A local secondary index
    """

    @classmethod
    def _update_model_schema(cls, schema: ModelSchema) -> None:
        index_schema = cls._get_schema()
        schema['local_secondary_indexes'].append(index_schema)
        # With polymorphism, indexes can use the same attribute, e.g. index1 on (thread_id, created_at)
        # and index2 on (thread_id, updated_at). We need to deduplicate.
        for attr_def in index_schema['attribute_definitions']:
            if attr_def not in schema['attribute_definitions']:
                schema['attribute_definitions'].append(attr_def)



class Projection:
    """
    A class for presenting projections
    """
    projection_type: Any = None
    non_key_attributes: Any = None


class KeysOnlyProjection(Projection):
    """
    Keys only projection
    """
    projection_type = KEYS_ONLY


class IncludeProjection(Projection):
    """
    An INCLUDE projection
    """
    projection_type = INCLUDE

    def __init__(self, non_attr_keys: Optional[List[str]] = None) -> None:
        if not non_attr_keys:
            raise ValueError("The INCLUDE type projection requires a list of string attribute names")
        self.non_key_attributes = non_attr_keys


class AllProjection(Projection):
    """
    An ALL projection
    """
    projection_type = ALL



"""
PynamoDB exceptions
"""
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing_extensions import Literal

import botocore.exceptions


class PynamoDBException(Exception):
    """
    Base class for all PynamoDB exceptions.
    """

    msg: str

    def __init__(self, msg: Optional[str] = None, cause: Optional[Exception] = None) -> None:
        self.msg = msg if msg is not None else self.msg
        self.cause = cause
        super(PynamoDBException, self).__init__(self.msg)

    @property
    def cause_response_code(self) -> Optional[str]:
        """
        The DynamoDB response code such as:

        - ``ConditionalCheckFailedException``
        - ``ProvisionedThroughputExceededException``
        - ``TransactionCanceledException``

        Inspect this value to determine the cause of the error and handle it.
        """
        return getattr(self.cause, 'response', {}).get('Error', {}).get('Code')

    @property
    def cause_response_message(self) -> Optional[str]:
        """
        The human-readable description of the error returned by DynamoDB.
        """
        return getattr(self.cause, 'response', {}).get('Error', {}).get('Message')


class PynamoDBConnectionError(PynamoDBException):
    """
    A base class for connection errors
    """
    msg = "Connection Error"


class DeleteError(PynamoDBConnectionError):
    """
    Raised when an error occurs deleting an item
    """
    msg = "Error deleting item"


class QueryError(PynamoDBConnectionError):
    """
    Raised when queries fail
    """
    msg = "Error performing query"


class ScanError(PynamoDBConnectionError):
    """
    Raised when a scan operation fails
    """
    msg = "Error performing scan"


class PutError(PynamoDBConnectionError):
    """
    Raised when an item fails to be created
    """
    msg = "Error putting item"


class UpdateError(PynamoDBConnectionError):
    """
    Raised when an item fails to be updated
    """
    msg = "Error updating item"


class GetError(PynamoDBConnectionError):
    """
    Raised when an item fails to be retrieved
    """
    msg = "Error getting item"


class TableError(PynamoDBConnectionError):
    """
    An error involving a dynamodb table operation
    """
    msg = "Error performing a table operation"


class DoesNotExist(PynamoDBException):
    """
    Raised when an item queried does not exist
    """
    msg = "Item does not exist"


class TableDoesNotExist(PynamoDBException):
    """
    Raised when an operation is attempted on a table that doesn't exist
    """
    def __init__(self, table_name: str) -> None:
        msg = "Table does not exist: `{}`".format(table_name)
        super(TableDoesNotExist, self).__init__(msg)


@dataclass
class CancellationReason:
    """
    A reason for a transaction cancellation.
    """
    code: Literal[
        'ConditionalCheckFailed',
        'ItemCollectionSizeLimitExceeded',
        'TransactionConflict',
        'ProvisionedThroughputExceeded',
        'ThrottlingError',
        'ValidationError',
    ]
    message: Optional[str] = None


class TransactWriteError(PynamoDBException):
    """
    Raised when a :class:`~pynamodb.transactions.TransactWrite` operation fails.
    """

    @property
    def cancellation_reasons(self) -> List[Optional[CancellationReason]]:
        """
        When :attr:`.cause_response_code` is ``TransactionCanceledException``, this property lists
        cancellation reasons in the same order as the transaction items (one-to-one).
        Items which were not part of the reason for cancellation would have :code:`None` as the value.

        For a list of possible cancellation reasons and their semantics,
        see `TransactWriteItems`_ in the AWS documentation.

        .. _TransactWriteItems: https://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_TransactWriteItems.html
        """
        if not isinstance(self.cause, VerboseClientError):
            return []
        return self.cause.cancellation_reasons


class TransactGetError(PynamoDBException):
    """
    Raised when a :class:`~pynamodb.transactions.TransactGet` operation fails.
    """
    @property
    def cancellation_reasons(self) -> List[Optional[CancellationReason]]:
        """
        When :attr:`.cause_response_code` is ``TransactionCanceledException``, this property lists
        cancellation reasons in the same order as the transaction items (one-to-one).
        Items which were not part of the reason for cancellation would have :code:`None` as the value.

        For a list of possible cancellation reasons and their semantics,
        see `TransactGetItems`_ in the AWS documentation.

        .. _TransactGetItems: https://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_TransactGetItems.html
        """
        if not isinstance(self.cause, VerboseClientError):
            return []
        return self.cause.cancellation_reasons


class InvalidStateError(PynamoDBException):
    """
    Raises when the internal state of an operation context is invalid.
    """
    msg = "Operation in invalid state"


class AttributeDeserializationError(TypeError):
    """
    Raised when attribute type is invalid during deserialization.
    """
    def __init__(self, attr_name: str, attr_type: str):
        msg = "Cannot deserialize '{}' attribute from type: {}".format(attr_name, attr_type)
        super(AttributeDeserializationError, self).__init__(msg)


class AttributeNullError(ValueError):
    """
    Raised when an attribute which is not nullable (:code:`null=False`) is unset during serialization.
    """

    def __init__(self, attr_name: str) -> None:
        self.attr_path = attr_name

    def __str__(self):
        return f"Attribute '{self.attr_path}' cannot be None"

    def prepend_path(self, attr_name: str) -> None:
        self.attr_path = attr_name + '.' + self.attr_path


class VerboseClientError(botocore.exceptions.ClientError):
    def __init__(
        self,
        error_response: Dict[str, Any],
        operation_name: str,
        verbose_properties: Optional[Any] = None,
        *,
        cancellation_reasons: Iterable[Optional[CancellationReason]] = (),
    ) -> None:
        """
        Like ClientError, but with a verbose message.

        :param error_response: Error response in shape expected by ClientError.
        :param operation_name: The name of the operation that failed.
        :param verbose_properties: A dict of properties to include in the verbose message.
        :param cancellation_reasons: For `TransactionCanceledException` error code,
          a list of cancellation reasons in the same order as the transaction's items (one to one).
          For items which were not a reason for the transaction cancellation, :code:`None` will be the value.
        """
        if not verbose_properties:
            verbose_properties = {}

        self.MSG_TEMPLATE = (
            'An error occurred ({{error_code}}) on request ({request_id}) '
            'on table ({table_name}) when calling the {{operation_name}} '
            'operation: {{error_message}}'
        ).format(request_id=verbose_properties.get('request_id'), table_name=verbose_properties.get('table_name'))

        self.cancellation_reasons = list(cancellation_reasons)

        super(VerboseClientError, self).__init__(
            error_response,  # type:ignore[arg-type]  # in stubs: botocore.exceptions._ClientErrorResponseTypeDef
            operation_name,
        )



from typing import Any
from typing import Dict


# match dynamo function syntax: size(path)
def size(path):
    from pynamodb.expressions.operand import _Size
    return _Size(path)


class Condition(object):
    format_string: str = ''

    def __init__(self, operator: str, *values) -> None:
        self.operator = operator
        self.values = values

    def __eq__(self, other: Any) -> bool:
        return (
            type(self) is type(other)
            and self.operator == other.operator
            and len(self.values) == len(other.values)
            and all(v1._equals_to(v2) for v1, v2 in zip(self.values, other.values))
        )

    def serialize(self, placeholder_names: Dict[str, str], expression_attribute_values: Dict[str, str]) -> str:
        values = [value.serialize(placeholder_names, expression_attribute_values) for value in self.values]
        return self.format_string.format(*values, operator=self.operator)

    def __and__(self, other):
        if not isinstance(other, Condition):
            raise TypeError("unsupported operand type(s) for &: '{}' and '{}'"
                            .format(self.__class__.__name__, other.__class__.__name__))
        return And(self, other)

    def __rand__(self, other):
        # special case 'None & condition' to enable better syntax for chaining
        if other is not None:
            raise TypeError("unsupported operand type(s) for &: '{}' and '{}'"
                            .format(other.__class__.__name__, self.__class__.__name__))
        return self

    def __or__(self, other):
        if not isinstance(other, Condition):
            raise TypeError("unsupported operand type(s) for |: '{}' and '{}'"
                            .format(self.__class__.__name__, other.__class__.__name__))
        return Or(self, other)

    def __invert__(self):
        return Not(self)

    def __repr__(self) -> str:
        values = [str(value) for value in self.values]
        return self.format_string.format(*values, operator=self.operator)

    def __bool__(self):
        # Prevent users from accidentally comparing the condition object instead of the attribute instance
        raise TypeError("unsupported operand type(s) for bool: {}".format(self.__class__.__name__))


class Comparison(Condition):
    format_string = '{0} {operator} {1}'

    def __init__(self, operator, lhs, rhs):
        if operator not in ['=', '<>', '<', '<=', '>', '>=']:
            raise ValueError("{0} is not a valid comparison operator: {0}".format(operator))
        super().__init__(operator, lhs, rhs)


class Between(Condition):
    format_string = '{0} {operator} {1} AND {2}'

    def __init__(self, path, lower, upper):
        super().__init__('BETWEEN', path, lower, upper)


class In(Condition):
    def __init__(self, path, *values):
        super().__init__('IN', path, *values)
        list_format = ', '.join('{' + str(i + 1) + '}' for i in range(len(values)))
        self.format_string = '{0} {operator} (' + list_format + ')'


class Exists(Condition):
    format_string = '{operator} ({0})'

    def __init__(self, path):
        super().__init__('attribute_exists', path)


class NotExists(Condition):
    format_string = '{operator} ({0})'

    def __init__(self, path):
        super().__init__('attribute_not_exists', path)


class IsType(Condition):
    format_string = '{operator} ({0}, {1})'

    def __init__(self, path, attr_type):
        super().__init__('attribute_type', path, attr_type)


class BeginsWith(Condition):
    format_string = '{operator} ({0}, {1})'

    def __init__(self, path, prefix):
        super().__init__('begins_with', path, prefix)


class Contains(Condition):
    format_string = '{operator} ({0}, {1})'

    def __init__(self, path, operand):
        super().__init__('contains', path, operand)


class And(Condition):
    format_string = '({0} {operator} {1})'

    def __init__(self, condition1: Condition, condition2: Condition) -> None:
        super().__init__('AND', condition1, condition2)


class Or(Condition):
    format_string = '({0} {operator} {1})'

    def __init__(self, condition1: Condition, condition2: Condition) -> None:
        super().__init__('OR', condition1, condition2)


class Not(Condition):
    format_string = '({operator} {0})'

    def __init__(self, condition: Condition) -> None:
        super().__init__('NOT', condition)



from typing import Any, Dict, List, Optional, Union
from typing import TYPE_CHECKING

from pynamodb.constants import (
    ATTRIBUTE_TYPES, BINARY_SET, LIST, MAP, NUMBER, NUMBER_SET, STRING, STRING_SET
)
from pynamodb.expressions.condition import (
    BeginsWith, Between, Comparison, Contains, Exists, In, IsType, NotExists
)
from pynamodb.expressions.update import (
    AddAction, DeleteAction, RemoveAction, SetAction
)
from pynamodb.expressions.util import get_path_segments, get_value_placeholder, substitute_names

if TYPE_CHECKING:
    from pynamodb.attributes import Attribute


class _Operand:
    """
    Operand is the base class for objects that can be operands in Condition and Update Expressions.
    """
    format_string = ''
    attr_type: Optional[str] = None

    def __init__(self, *values: Any) -> None:
        self.values = values

    def __repr__(self) -> str:
        return self.format_string.format(*self.values)

    def _equals_to(self, other: Any) -> bool:
        return (
            type(self) is type(other)
            and self.values == other.values
        )

    def serialize(self, placeholder_names: Dict[str, str], expression_attribute_values: Dict[str, str]) -> str:
        values = [self._serialize_value(value, placeholder_names, expression_attribute_values) for value in self.values]
        return self.format_string.format(*values)

    def _serialize_value(self, value, placeholder_names, expression_attribute_values):
        return value.serialize(placeholder_names, expression_attribute_values)

    def _to_operand(self, value: Union['_Operand', 'Attribute', Any]):
        if isinstance(value, _Operand):
            return value
        from pynamodb.attributes import Attribute, MapAttribute  # prevent circular import -- Attribute imports Path
        if isinstance(value, MapAttribute) and value._is_attribute_container():
            return self._to_value(value)
        return Path(value) if isinstance(value, Attribute) else self._to_value(value)

    def _to_value(self, value):
        return Value(value)

    def _type_check(self, *types):
        if self.attr_type and self.attr_type not in types:
            raise ValueError("The data type of '{}' must be one of {}".format(self, list(types)))


class _ConditionOperand(_Operand):
    """
    A base class for Operands that can be used in Condition Expression comparisons.
    """

    def __eq__(self, other: Any) -> Comparison:  # type: ignore[override]
        return Comparison('=', self, self._to_operand(other))

    def __ne__(self, other: Any) -> Comparison:  # type: ignore[override]
        return Comparison('<>', self, self._to_operand(other))

    def __lt__(self, other: Any) -> Comparison:
        return Comparison('<', self, self._to_operand(other))

    def __le__(self, other: Any) -> Comparison:
        return Comparison('<=', self, self._to_operand(other))

    def __gt__(self, other: Any) -> Comparison:
        return Comparison('>', self, self._to_operand(other))

    def __ge__(self, other: Any) -> Comparison:
        return Comparison('>=', self, self._to_operand(other))

    def between(self, lower: Any, upper: Any) -> Between:
        return Between(self, self._to_operand(lower), self._to_operand(upper))

    def is_in(self, *values: Any) -> In:
        op_values = [self._to_operand(value) for value in values]
        return In(self, *op_values)


class _NumericOperand(_Operand):
    """
    A base class for Operands that can be used in the increment and decrement SET update actions.
    """

    def __add__(self, other):
        return _Increment(self, self._to_operand(other))

    def __radd__(self, other):
        return _Increment(self._to_operand(other), self)

    def __sub__(self, other):
        return _Decrement(self, self._to_operand(other))

    def __rsub__(self, other):
        return _Decrement(self._to_operand(other), self)


class _ListAppendOperand(_Operand):
    """
    A base class for Operands that can be used in the list_append function for the SET update action.
    """

    def append(self, other: Any) -> '_ListAppend':
        return _ListAppend(self, self._to_operand(other))

    def prepend(self, other: Any) -> '_ListAppend':
        return _ListAppend(self._to_operand(other), self)


class _Size(_ConditionOperand):
    """
    Size is a special operand that represents the result of calling the 'size' function on a Path operand.
    """
    format_string = 'size ({0})'
    attr_type = NUMBER

    def __init__(self, path: Union['Path', 'Attribute', str, List[str]]) -> None:
        if not isinstance(path, Path):
            path = Path(path)
        super(_Size, self).__init__(path)

    def _to_operand(self, value):
        operand = super(_Size, self)._to_operand(value)
        operand._type_check(NUMBER)
        return operand


class _Increment(_Operand):
    """
    Increment is a special operand that represents an increment SET update action.
    """
    format_string = '{0} + {1}'
    attr_type = NUMBER

    def __init__(self, lhs: '_Operand', rhs: '_Operand') -> None:
        lhs._type_check(NUMBER)
        rhs._type_check(NUMBER)
        super(_Increment, self).__init__(lhs, rhs)


class _Decrement(_Operand):
    """
    Decrement is a special operand that represents an decrement SET update action.
    """
    format_string = '{0} - {1}'
    attr_type = NUMBER

    def __init__(self, lhs: _Operand, rhs: _Operand) -> None:
        lhs._type_check(NUMBER)
        rhs._type_check(NUMBER)
        super(_Decrement, self).__init__(lhs, rhs)


class _ListAppend(_Operand):
    """
    ListAppend is a special operand that represents the list_append function for the SET update action.
    """
    format_string = 'list_append ({0}, {1})'
    attr_type = LIST

    def __init__(self, list1: _Operand, list2: _Operand):
        list1._type_check(LIST)
        list2._type_check(LIST)
        super(_ListAppend, self).__init__(list1, list2)


class _IfNotExists(_NumericOperand, _ListAppendOperand):
    """
    IfNotExists is a special operand that represents the if_not_exists function for the SET update action.
    """
    format_string = 'if_not_exists ({0}, {1})'

    def __init__(self, path: _Operand, value: Any) -> None:
        self.attr_type = path.attr_type or value.attr_type
        if self.attr_type != value.attr_type:
            # path and value have conflicting types -- defer any type checks to DynamoDB
            self.attr_type = None
        super(_IfNotExists, self).__init__(path, value)


class Value(_NumericOperand, _ListAppendOperand, _ConditionOperand):
    """
    Value is an operand that represents an attribute value.
    """
    format_string = '{0}'

    def __init__(self, value: Any, attribute: Optional['Attribute'] = None) -> None:
        # Check to see if value is already serialized
        if isinstance(value, dict) and len(value) == 1 and list(value.keys())[0] in ATTRIBUTE_TYPES:
            (self.attr_type, value), = value.items()
        elif value is None:
            (self.attr_type, value) = Value.__serialize(value)
        else:
            (self.attr_type, value) = Value.__serialize(value, attribute)
        super(Value, self).__init__({self.attr_type: value})

    def _equals_to(self, other: Any) -> bool:
        return (
            super()._equals_to(other)
            and self.attr_type == other.attr_type
        )

    @property
    def value(self):
        return self.values[0]

    def _serialize_value(self, value, placeholder_names, expression_attribute_values):
        return get_value_placeholder(value, expression_attribute_values)

    @staticmethod
    def __serialize(value, attribute=None):
        if attribute is None:
            return Value.__serialize_based_on_type(value)
        if attribute.attr_type == LIST and not isinstance(value, list):
            # List attributes assume the values to be serialized are lists.
            (attr_type, attr_value), = attribute.serialize([value])[0].items()
            return attr_type, attr_value
        if attribute.attr_type == MAP and not isinstance(value, dict):
            # Map attributes assume the values to be serialized are maps.
            return Value.__serialize_based_on_type(value)
        return attribute.attr_type, attribute.serialize(value)

    @staticmethod
    def __serialize_based_on_type(value):
        from pynamodb.attributes import _get_class_for_serialize
        attr_class = _get_class_for_serialize(value)
        return attr_class.attr_type, attr_class.serialize(value)


class Path(_NumericOperand, _ListAppendOperand, _ConditionOperand):
    """
    Path is an operand that represents either an attribute name or document path.
    """
    format_string = '{0}'

    def __init__(self, attribute_or_path: Union['Attribute', str, List[str]]) -> None:
        from pynamodb.attributes import Attribute  # prevent circular import -- Attribute imports Path
        path: Union[str, List[str]]
        if isinstance(attribute_or_path, Attribute):
            self.attribute = attribute_or_path
            self.attr_type = attribute_or_path.attr_type
            path = attribute_or_path.attr_path
        else:
            self.attribute = None
            self.attr_type = None
            path = attribute_or_path
        if not path:
            raise ValueError("path cannot be empty")
        super(Path, self).__init__(get_path_segments(path))

    def _equals_to(self, other: Any) -> bool:
        return (
            super()._equals_to(other)
            and self.attr_type == other.attr_type
            # 'is' to equate class identity (rather than trigger Attribute.__eq__)
            and self.attribute is other.attribute
        )

    @property
    def path(self) -> List[str]:
        return self.values[0]

    def __iter__(self):
        # Because we define __getitem__ Path is considered an iterable
        raise TypeError("'{}' object is not iterable".format(self.__class__.__name__))

    def __getitem__(self, item: Union[int, str]) -> 'Path':
        # The __getitem__ call returns a new Path instance without any attribute set.
        # This is intended since the nested element is not the same attribute as ``self``.
        if self.attribute and self.attribute.attr_type not in [LIST, MAP]:
            raise TypeError("'{}' object has no attribute __getitem__".format(self.attribute.__class__.__name__))
        if self.attr_type == LIST and not isinstance(item, int):
            raise TypeError("list indices must be integers, not {}".format(type(item).__name__))
        if self.attr_type == MAP and not isinstance(item, str):
            raise TypeError("map attributes must be strings, not {}".format(type(item).__name__))
        if isinstance(item, int):
            # list dereference operator
            element_path = Path(self.path)  # copy the document path before indexing last element
            element_path.path[-1] = '{}[{}]'.format(self.path[-1], item)
            return element_path
        if isinstance(item, str):
            # map dereference operator
            return Path(self.path + [item])
        raise TypeError("item must be an integer or string, not {}".format(type(item).__name__))

    def __or__(self, other):
        return _IfNotExists(self, self._to_operand(other))

    def set(self, value: Any) -> SetAction:
        # Returns an update action that sets this attribute to the given value
        return SetAction(self, self._to_operand(value))

    def remove(self) -> RemoveAction:
        # Returns an update action that removes this attribute from the item
        return RemoveAction(self)

    def add(self, *values: Any) -> AddAction:
        # Returns an update action that appends the given values to a set or mathematically adds a value to a number
        value = values[0] if len(values) == 1 else values
        return AddAction(self, self._to_operand(value))

    def delete(self, *values: Any) -> DeleteAction:
        # Returns an update action that removes the given values from a set attribute
        value = values[0] if len(values) == 1 else values
        return DeleteAction(self, self._to_operand(value))

    def exists(self) -> Exists:
        return Exists(self)

    def does_not_exist(self) -> NotExists:
        return NotExists(self)

    def is_type(self, attr_type: str) -> IsType:
        if attr_type not in ATTRIBUTE_TYPES:
            raise ValueError("{} is not a valid attribute type. Must be one of {}".format(
                attr_type, ATTRIBUTE_TYPES))
        return IsType(self, Value(attr_type))

    def startswith(self, prefix: str) -> BeginsWith:
        # A 'pythonic' replacement for begins_with to match string behavior (e.g. "foo".startswith("f"))
        operand = self._to_operand(prefix)
        operand._type_check(STRING)
        return BeginsWith(self, operand)

    def contains(self, item: Any) -> Contains:
        if self.attribute and self.attribute.attr_type in [BINARY_SET, NUMBER_SET, STRING_SET]:
            # Set attributes assume the values to be serialized are sets.
            (attr_type, attr_value), = self._to_value([item]).value.items()
            item = {attr_type[0]: attr_value[0]}
        return Contains(self, self._to_operand(item))

    def _serialize_value(self, value, placeholder_names, expression_attribute_values):
        return substitute_names(value, placeholder_names)

    def _to_value(self, value: Any) -> Value:
        return Value(value, attribute=self.attribute)

    def __str__(self) -> str:
        # Quote the path to illustrate that any dot characters are not dereference operators.
        quoted_path = [self._quote_path(segment) if '.' in segment else segment for segment in self.path]
        return '.'.join(quoted_path)

    def __repr__(self) -> str:
        return "Path({})".format(self.path)

    @staticmethod
    def _quote_path(path: str) -> str:
        path, sep, rem = path.partition('[')
        return repr(path) + sep + rem



import importlib.util
import logging
import os
import warnings
from os import getenv

from typing import Any, Optional, Mapping, ClassVar

log = logging.getLogger(__name__)

default_settings_dict = {
    'connect_timeout_seconds': 15,
    'read_timeout_seconds': 30,
    'max_retry_attempts': 3,
    'base_backoff_ms': 25,
    'region': None,
    'max_pool_connections': 10,
    'extra_headers': None,
    'dax_write_endpoints': [],
    'dax_read_endpoints': []
}

OVERRIDE_SETTINGS_PATH = getenv('PYNAMODB_CONFIG', '/etc/pynamodb/global_default_settings.py')


def _load_module(name, path):
    # https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
    spec = importlib.util.spec_from_file_location(name, path)
    module = importlib.util.module_from_spec(spec)  # type: ignore
    spec.loader.exec_module(module)  # type: ignore
    return module


override_settings = {}
if os.path.isfile(OVERRIDE_SETTINGS_PATH):
    override_settings = _load_module('__pynamodb_override_settings__', OVERRIDE_SETTINGS_PATH)
    if hasattr(override_settings, 'session_cls') or hasattr(override_settings, 'request_timeout_seconds'):
        warnings.warn("The `session_cls` and `request_timeout_second` options are no longer supported")
    log.info('Override settings for pynamo available {}'.format(OVERRIDE_SETTINGS_PATH))
else:
    log.info('Override settings for pynamo not available {}'.format(OVERRIDE_SETTINGS_PATH))
    log.info('Using Default settings value')


def get_settings_value(key: str) -> Any:
    """
    Fetches the value from the override file.
    If the value is not present, then tries to fetch the values from constants.py
    """
    if hasattr(override_settings, key):
        return getattr(override_settings, key)

    if key in default_settings_dict:
        return default_settings_dict[key]

    return None


class OperationSettings:
    """
    Settings applicable to an individual operation.
    When set, the settings in this object supersede the global and model settings.
    """
    default: ClassVar['OperationSettings']

    def __init__(self, *, extra_headers: Optional[Mapping[str, Optional[str]]] = None) -> None:
        """
        Initializes operation settings.
        :param extra_headers: if set, extra headers to add to the HTTP request. The headers are merged
         on top of extra headers derived from settings or models' Meta classes. To delete a header, set its value
         to `None`.
        """
        self.extra_headers = extra_headers


OperationSettings.default = OperationSettings()



"""
Mock response
"""
from urllib3 import HTTPResponse


class MockResponse(HTTPResponse):
    """
    A class for mocked responses
    """
    def __init__(self, status_code=None, content='Empty'):
        super(MockResponse, self).__init__()
        self.status_code = status_code
        self._content = content
        self.reason = 'Test Response'


class HttpOK(MockResponse):
    """
    A response that returns status code 200
    """
    def __init__(self, content=None):
        super(HttpOK, self).__init__(status_code=200, content=content)



from typing_extensions import assert_type


def test_transact_write() -> None:
    from pynamodb.transactions import TransactWrite
    with TransactWrite() as tx:
        assert_type(tx, TransactWrite)


def test_transact_get() -> None:
    from pynamodb.transactions import TransactGet
    from pynamodb.models import Model, _ModelFuture

    class FirstModel(Model):
        pass

    class SecondModel(Model):
        pass

    with TransactGet() as tx:
        assert_type(tx, TransactGet)
        assert_type(tx.get(FirstModel, "pk"), _ModelFuture[FirstModel])
        assert_type(tx.get(SecondModel, "pk"), _ModelFuture[SecondModel])

        second_model_instance_future = tx.get(SecondModel, "pk")

    assert_type(second_model_instance_future.get(), SecondModel)
    _first_model_instance: FirstModel = second_model_instance_future.get()  # type:ignore[assignment]



import json
from base64 import b64decode
from base64 import b64encode
from typing import Any
from typing import Dict

from pynamodb.constants import BINARY
from pynamodb.constants import BINARY_SET
from pynamodb.constants import BOOLEAN
from pynamodb.constants import LIST
from pynamodb.constants import MAP
from pynamodb.constants import NULL
from pynamodb.constants import NUMBER
from pynamodb.constants import NUMBER_SET
from pynamodb.constants import STRING
from pynamodb.constants import STRING_SET


def attr_value_to_simple_dict(attribute_value: Dict[str, Any], force: bool) -> Any:
    attr_type, attr_value = next(iter(attribute_value.items()))
    if attr_type == LIST:
        return [attr_value_to_simple_dict(v, force) for v in attr_value]
    if attr_type == MAP:
        return {k: attr_value_to_simple_dict(v, force) for k, v in attr_value.items()}
    if attr_type == NULL:
        return None
    if attr_type == BOOLEAN:
        return attr_value
    if attr_type == STRING:
        return attr_value
    if attr_type == NUMBER:
        return json.loads(attr_value)
    if attr_type == BINARY:
        if force:
            return b64encode(attr_value).decode()
        raise ValueError("Binary attributes are not supported")
    if attr_type == BINARY_SET:
        if force:
            return [b64encode(v).decode() for v in attr_value]
        raise ValueError("Binary set attributes are not supported")
    if attr_type == STRING_SET:
        if force:
            return attr_value
        raise ValueError("String set attributes are not supported")
    if attr_type == NUMBER_SET:
        if force:
            return [json.loads(v) for v in attr_value]
        raise ValueError("Number set attributes are not supported")
    raise ValueError("Unknown attribute type: {}".format(attr_type))


def simple_dict_to_attr_value(value: Any) -> Dict[str, Any]:
    if value is None:
        return {NULL: True}
    if value is True or value is False:
        return {BOOLEAN: value}
    if isinstance(value, (int, float)):
        return {NUMBER: json.dumps(value)}
    if isinstance(value, str):
        return {STRING: value}
    if isinstance(value, list):
        return {LIST: [simple_dict_to_attr_value(v) for v in value]}
    if isinstance(value, dict):
        return {MAP: {k: simple_dict_to_attr_value(v) for k, v in value.items()}}
    raise ValueError("Unknown value type: {}".format(type(value).__name__))


def _b64encode(b: bytes) -> str:
    return b64encode(b).decode()


def bin_encode_attr(attr: Dict[str, Any]) -> None:
    if BINARY in attr:
        attr[BINARY] = _b64encode(attr[BINARY])
    elif BINARY_SET in attr:
        attr[BINARY_SET] = [_b64encode(v) for v in attr[BINARY_SET]]
    elif MAP in attr:
        for sub_attr in attr[MAP].values():
            bin_encode_attr(sub_attr)
    elif LIST in attr:
        for sub_attr in attr[LIST]:
            bin_encode_attr(sub_attr)


def bin_decode_attr(attr: Dict[str, Any]) -> None:
    if BINARY in attr:
        attr[BINARY] = b64decode(attr[BINARY])
    elif BINARY_SET in attr:
        attr[BINARY_SET] = [b64decode(v) for v in attr[BINARY_SET]]
    elif MAP in attr:
        for sub_attr in attr[MAP].values():
            bin_decode_attr(sub_attr)
    elif LIST in attr:
        for sub_attr in attr[LIST]:
            bin_decode_attr(sub_attr)



import pytest
from pynamodb.attributes import NumberAttribute, UnicodeAttribute, VersionAttribute

from pynamodb.connection import Connection
from pynamodb.connection.base import MetaTable
from pynamodb.constants import TABLE_KEY
from pynamodb.transactions import Transaction, TransactGet, TransactWrite
from pynamodb.models import Model
from tests.test_base_connection import PATCH_METHOD

from unittest.mock import patch


class MockModel(Model):
    class Meta:
        table_name = 'mock'

    mock_hash = NumberAttribute(hash_key=True)
    mock_range = NumberAttribute(range_key=True)
    mock_toot = UnicodeAttribute(null=True)
    mock_version = VersionAttribute()


MOCK_TABLE_DESCRIPTOR = {
    "Table": {
        "TableName": "mock",
        "KeySchema": [
            {
                "AttributeName": "mock_hash",
                "KeyType": "HASH"
            },
            {
                "AttributeName": "mock_range",
                "KeyType": "RANGE"
            }
        ],
        "AttributeDefinitions": [
            {
                "AttributeName": "mock_hash",
                "AttributeType": "N"
            },
            {
                "AttributeName": "mock_range",
                "AttributeType": "N"
            }
        ]
    }
}


class TestTransaction:

    def test_commit__not_implemented(self):
        t = Transaction(connection=Connection())
        with pytest.raises(NotImplementedError):
            t._commit()


class TestTransactGet:

    def test_commit(self, mocker):
        connection = Connection()
        connection.add_meta_table(MetaTable(MOCK_TABLE_DESCRIPTOR[TABLE_KEY]))

        mock_connection_transact_get = mocker.patch.object(connection, 'transact_get_items')

        with TransactGet(connection=connection) as t:
            t.get(MockModel, 1, 2)

        mock_connection_transact_get.assert_called_once_with(
            get_items=[{'Key': {'mock_hash': {'N': '1'}, 'mock_range': {'N': '2'}}, 'TableName': 'mock'}],
            return_consumed_capacity=None
        )


class TestTransactWrite:

    def test_condition_check__no_condition(self):
        with pytest.raises(TypeError):
            with TransactWrite(connection=Connection()) as transaction:
                transaction.condition_check(MockModel, hash_key=1, condition=None)

    def test_commit(self, mocker):
        connection = Connection()
        mock_connection_transact_write = mocker.patch.object(connection, 'transact_write_items')
        with TransactWrite(connection=connection) as t:
            t.condition_check(MockModel, 1, 3, condition=(MockModel.mock_hash.does_not_exist()))
            t.delete(MockModel(2, 4))
            t.save(MockModel(3, 5))
            t.update(MockModel(4, 6), actions=[MockModel.mock_toot.set('hello')], return_values='ALL_OLD')

        expected_condition_checks = [{
            'ConditionExpression': 'attribute_not_exists (#0)',
            'ExpressionAttributeNames': {'#0': 'mock_hash'},
            'Key': {'mock_hash': {'N': '1'}, 'mock_range': {'N': '3'}},
            'TableName': 'mock'}
        ]
        expected_deletes = [{
            'ConditionExpression': 'attribute_not_exists (#0)',
            'ExpressionAttributeNames': {'#0': 'mock_version'},
            'Key': {'mock_hash': {'N': '2'}, 'mock_range': {'N': '4'}},
            'TableName': 'mock'
        }]
        expected_puts = [{
            'ConditionExpression': 'attribute_not_exists (#0)',
            'ExpressionAttributeNames': {'#0': 'mock_version'},
            'Item': {'mock_hash': {'N': '3'}, 'mock_range': {'N': '5'}, 'mock_version': {'N': '1'}},
            'TableName': 'mock'
        }]
        expected_updates = [{
            'ConditionExpression': 'attribute_not_exists (#0)',
            'TableName': 'mock',
            'Key': {'mock_hash': {'N': '4'}, 'mock_range': {'N': '6'}},
            'ReturnValuesOnConditionCheckFailure': 'ALL_OLD',
            'UpdateExpression': 'SET #1 = :0, #0 = :1',
            'ExpressionAttributeNames': {'#0': 'mock_version', '#1': 'mock_toot'},
            'ExpressionAttributeValues': {':0': {'S': 'hello'}, ':1': {'N': '1'}}
        }]
        mock_connection_transact_write.assert_called_once_with(
            condition_check_items=expected_condition_checks,
            delete_items=expected_deletes,
            put_items=expected_puts,
            update_items=expected_updates,
            client_request_token=None,
            return_consumed_capacity=None,
            return_item_collection_metrics=None
        )
